from ..console import GREEN
from ..console import RED
from ..console import Console
from ..database import Category
from ..database import Project
from ..gui import default_text_input
from ..gui import number_input
from ..gui import text_input
from ..util import regex_factory
from .util import db_session


@db_session
def new_category(conf):
    category_text = text_input("Category")

    Console.newline()
    Console.newline()

    try:
        _category = Category(name=category_text)
        conf.db.add(_category)
        conf.db.commit()

    except Exception:
        Console.write(RED)
        Console.writeline("Update error. Category was not saved.")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write(f"New Category {category_text} saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def new_project(conf):
    subject_text = text_input("Project/Incident")

    parse_incident_id = regex_factory(
        (
            r"^.*(?P<type>WO|INC|CRQ|FY\d\d)-?"
            r"0*(?P<number>[0-9]\d+) "
            r"(?P<description>.+)$"
        )
    )

    incident_id = parse_incident_id(subject_text) or {}

    Console.newline()

    prj_type = default_text_input("Type", incident_id.get("type", "WO"))
    prj_num = number_input("Number", incident_id.get("number", ""))
    prj_desc = default_text_input("Desc", incident_id.get("description", ""))

    Console.newline()

    try:
        project = Project(type=prj_type, number=prj_num, name=prj_desc)
        conf.db.add(project)
        conf.db.commit()
    except Exception as ex:
        Console.write(RED)
        Console.writeline("Update error. Incicent was not saved.")
        Console.writeline(f"(Error: {ex})")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write("New incident saved!")
    Console.clear_formating()
    Console.newline()



from datetime import datetime
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy import create_engine
from sqlalchemy import desc
from sqlalchemy import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker
from ttmr.time import Date


Base = declarative_base()


def init(dbfile):
    engine = create_engine(f"sqlite:///{dbfile}")
    Base.metadata.create_all(bind=engine)
    return sessionmaker(bind=engine)


class BaseTable:  # pylint: disable=R0903
    id = Column(Integer, primary_key=True, autoincrement=True)
    active = Column(Boolean(), default=True)
    created = Column(DateTime(timezone=False), default=datetime.now)
    modified = Column(
        DateTime(timezone=False), default=datetime.now, onupdate=datetime.now
    )

    @classmethod
    def get_all(cls, db):
        return db.query(cls).all()

    @classmethod
    def get_active(cls, db):
        # pylint: disable=C0121
        return db.query(cls).filter(cls.active == True).all()  # noqa


FMTS = {
    "INC": "{}{:0>8}",
    "WO": "{}{:0>9}",
    "FY16": "{}-{:0>6}",
    "FY17": "{}-{:0>6}",
    "FY18": "{}-{:0>6}",
    "FY19": "{}-{:0>6}",
    "FY20": "{}-{:0>6}",
    "FY21": "{}-{:0>6}",
    "FY22": "{}-{:0>6}",
    "": "{}-{:0>8}",
}


class Project(Base, BaseTable):
    __tablename__ = "project"

    type = Column(String(25), nullable=False)
    number = Column(Integer, nullable=False, unique=True)
    name = Column(String(100), nullable=False)
    persistant = Column(Boolean(), default=False)

    @property
    def identifier(self):
        return FMTS.get(self.type, FMTS[""]).format(self.type, self.number)


class Category(Base, BaseTable):
    __tablename__ = "category"
    name = Column(String(100), nullable=False, unique=True)


class Entry(Base, BaseTable):
    __tablename__ = "entry"
    category_id = Column(Integer, ForeignKey(Category.id))
    category = relationship(Category, lazy="immediate")
    project_id = Column(Integer, ForeignKey(Project.id))
    project = relationship(Project)
    note = Column(Text, default="")
    start_time = Column(DateTime(timezone=False), unique=True)
    end_time = Column(DateTime(timezone=False), unique=True)
    duration = Column(Integer)

    @property
    def is_completed(self):
        return self.end_time and self.end_time

    @property
    def minutes(self):
        return self.duration

    @property
    def hours(self):
        return round((self.duration / 60.0), 2)

    def start(self):
        self.start_time = Date.current()

    def end(self, end_time=None):
        if not self.start_time:
            raise Exception("You cannot end an entry without a start time.")
        if end_time:
            self.end_time = end_time
        else:
            self.end_time = Date.current()
        self.duration = self.calculate_duration()

    def calculate_duration(self):
        if not self.end_time or not self.end_time:
            raise Exception("You cannot end an entry without a start time.")

        elapsed_time = self.end_time - self.start_time
        return int(elapsed_time.total_seconds() / 60)

    def __str__(self):
        return (
            f"   Category: {self.category.name}\n"
            f"   Incident: {self.project.name}\n"
            f"       Note: {self.note}\n"
            f" Start time: {self.start_time}\n"
            f"   End time: {self.end_time}\n"
            f"   Duration: {self.minutes}\n"
        )

    @classmethod
    def get_day_start_entry(cls, hour=7, minute=0):
        entry = cls(note="start of day")
        entry.start_time = end_time = Date.current().start_of_day(hour=hour, minute=minute)
        entry.end_time = end_time
        entry.duration = 0
        return entry

    @classmethod
    def get_last(cls, db, given_datetime):
        day = given_datetime.replace(hour=0, minute=0, second=0, microsecond=0)
        query = db.query(cls).filter(cls.start_time >= day)
        query = query.order_by(desc(cls.start_time))
        return query.first()

    @classmethod
    def current_week(cls, db, today=None):
        if not today:
            today = Date.current(date=today).start_of_week()
        bow = today.get_start_of_week()
        print(f"start of week: {bow}")
        query = db.query(cls).filter(cls.start_time >= bow)
        query = query.order_by(cls.start_time)
        return query.all()

    @classmethod
    def day_summary(cls, db, today=None):
        if not today:
            today = Date.current(date=today).start_of_day()
        query = db.query(
            func.min(cls.start_time),
            func.min(cls.start_time),
            func.max(cls.end_time),
            func.sum(cls.duration),
            (func.sum(cls.duration) / 60.0),
        )
        query = query.filter(cls.start_time >= today)
        return query.all()

    @classmethod
    def week_summary(cls, db, today=None):
        if not today:
            today = Date.current(date=today).get_start_of_week()
        query = db.query(Category.name, func.count(cls.id), func.sum(cls.duration))
        query = query.filter(cls.start_time >= today)
        return query.group_by(Category.name).all()



import os
from pathlib import Path
from types import SimpleNamespace
import tomlkit
from appdirs import AppDirs
from .time import Date


def assert_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def get_appdir():
    appdirs = AppDirs("ttmr", "mgemmill")
    user_data_dir = appdirs.user_data_dir
    assert_dir(user_data_dir)
    return user_data_dir


def get_config(args):
    cfg = SimpleNamespace()

    cfg.cli_args = args

    cfg.appdirs = appdirs = AppDirs("ttmr", "mgemmill")
    cfg.user_data_dir = app_dir = appdirs.user_data_dir
    cfg.user_cache_dir = cache_dir = Path(appdirs.user_cache_dir)

    assert_dir(app_dir)
    assert_dir(cache_dir)

    cfg.config_path = Path(app_dir, "ttmr.cfg").resolve()

    cfg.data = data = tomlkit.parse(cfg.config_path.read_text())
    cfg.db_path = Path(app_dir, data["ttmr"]["db"])
    #  cfg.timezone = timezone = data["ttmr"].get("timezone", "America/Vancouver")

    cfg.today = Date.current(args.get("--date"))

    hour, minute = [int(s) for s in data["ttmr"].get("start_time", "07:00").split(":")]

    cfg.day_start_time = cfg.today.replace(hour=hour, minute=minute)

    return cfg



from ttmr.time import Date
from ..console import Console
from ..database import Category
from ..database import Entry
from ..database import Project
from ..gui import bool_input
from ..gui import clear_line
from ..gui import date_input
from ..gui import display_alert
from ..gui import display_message
from ..gui import duration_input
from ..gui import option_input
from ..gui import show_timer
from ..gui import text_input
from .util import db_session
from .util import new_session


def time_entry(conf, ui, end_entry=False):
    """Creates a new entry. When doing so it ends the current entry
    (if there is one) and records its duration.

    When the new entry is created, the entries timer is displayed.

    The timer requires that there is a Day Start Entry.
    Day Start will have a start_time and end_time that are the same.

    There are 2 ways to create entries:

    1. as a ttmr start command:
        - this will open a new entry with a start time that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - the window remains open with a time counter that explains
          this is the count on the current timer.
        - when you close the window, the end time will be set.

    2. as a ttmr end command:
        - this will open a new entry with a start that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - this will also set the end time.
        - the window remains open with time counter

    """
    # fetch working data
    categories = Category.get_all(conf.db)
    projects = Project.get_all(conf.db)

    end_time = None
    duration = 0
    current_timestamp = Date.current()
    last_entry = Entry.get_last(conf.db, current_timestamp)

    if end_entry:

        if not last_entry:
            last_entry = Entry.get_day_start_entry()

        if last_entry.end_time is not None:
            current_timestamp = last_entry.end_time

    if not end_entry and last_entry and last_entry.duration == 0:
        # if there is a pre-exiting entry that is not closed
        current_timestamp = close_open_entry(conf.db, last_entry)

    # get inputs
    category, project, note, start_time = ui.fetch_entry_inputs(
        categories, projects, current_timestamp
    )

    # if this is an end entry...
    if end_entry:
        # TODO: it is a bit of a problem that we might be overriding the duration...
        # TODO: and that the duration is a captured value
        end_time, duration = ui.fetch_end_entry_inputs(start_time, current_timestamp)

    # create entry
    entry = Entry(
        category=category,
        project=project,
        note=note,
        start_time=start_time,
        end_time=end_time,
        duration=duration,
    )

    conf.db.add(entry)
    conf.db.commit()

    return entry


@db_session
def show_current_entry(conf):
    """Show the last or current entry and display it's timer.
    If there is no open entries, it shows nothing.

    """
    last_entry = Entry.get_last(conf.db, Date.current())

    if not last_entry or last_entry.duration != 0:
        display_alert(" There are no active entries!")
        return

    display_message(str(last_entry))

    EntryUI.post_entry_timer(conf, "", last_entry.start_time, last_entry.id)


@db_session
def start_time_entry(cfg):
    entry = time_entry(cfg, EntryUI, end_entry=False)
    EntryUI.post_entry_timer(cfg, "Current Project Timer:", entry.start_time, entry.id)


@db_session
def stop_time_entry(cfg):
    entry = time_entry(cfg, EntryUI, end_entry=True)
    EntryUI.post_entry_timer(cfg, "Time since last completed entry::", entry.end_time)


def close_open_entry(db, current_entry):

    clear_line()
    print("You have an open entry.")
    end_entry = bool_input("End the current entry?")
    if end_entry:
        current_entry.end()
        db.commit()

        Console.cursor_up()
        Console.cursor_to_left_margin()
        Console.clear_line()
        Console.cursor_up()
        Console.cursor_to_left_margin()
        Console.clear_line()

        return current_entry.end_time

    raise KeyboardInterrupt()


#  def edit_entry(conf):
#      if not conf.cli_curr_obj:
#          print("We don't have the selection feature built yet!")
#          return
#
#      with Database(conf.sqlite, conf.dt) as db:
#          last_entry = db.fetch_last_entry()
#          categories = db.fetch_category()
#          incidents = db.fetch_incidents()
#
#      # get inputs
#      entry_id = last_entry.entry_id
#      category = option_input("Category", categories, default=last_entry.category)
#      incident = option_input("Incident", incidents, default=last_entry.incident)
#
#      notes = text_input("Notes", default=last_entry.note)
#      timestamp = date_input("Date", conf.dt, default=last_entry.start_time)
#
#      do_update = bool_input("Proceed with update?")
#
#      if not do_update:
#          print("canceling update")
#          return
#
#      with Database(conf.sqlite, conf.dt) as db:
#          db.update_entry(entry_id, category, incident, notes, timestamp)


class EntryUI:
    @staticmethod
    def fetch_entry_inputs(categories, projects, last_end_time):
        category = option_input("Category", categories)
        project = option_input("Project", projects)
        note = text_input("Notes")
        start_time = date_input("Start Date", default=last_end_time)
        return category, project, note, start_time

    @staticmethod
    def fetch_end_entry_inputs(start_time, current_timestamp):
        end_time = date_input("End Date", default=current_timestamp)
        elapsed_time = 0
        elapsed_time = current_timestamp - start_time
        duration = duration_input(
            "Duration", str(int(elapsed_time.total_seconds() / 60))
        )
        return end_time, duration

    @staticmethod
    def post_entry_timer(cfg, msg, current_timestamp, current_entry_id=None):
        print()

        if current_entry_id:
            print(msg)
            try:
                show_timer(current_timestamp)
            except KeyboardInterrupt:
                clear_line()
                end_entry = bool_input("End the current entry?")
                if end_entry:
                    print("Closing current session...")
                    with new_session(cfg.db_path) as db:
                        current_entry = (
                            db.query(Entry)
                            .filter(Entry.id == current_entry_id)
                            .one_or_none()
                        )
                        current_entry.end()
                        db.commit()
        else:
            print(msg)
            show_timer(current_timestamp)



from datetime import datetime
from datetime import timedelta
from dateutil import parser

FMT_DATE = "{:%Y-%m-%d}"
FMT_DATETIME = "{:%Y-%m-%d %H:%M}"


class Date(datetime):
    def std_weekday(self):
        """
        Sunday = 1
        Saturday = 7

        """
        d = self.weekday() + 2
        if d == 8:
            return 1
        return d

    @classmethod
    def current(cls, date=None):
        if not date:
            return Date.from_datetime(datetime.now())
        if isinstance(date, str):
            return Date.from_datetime(parser.parse(date))
        raise Exception("Invalid date input!")

    @classmethod
    def from_datetime(cls, dt):
        return Date(
            dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond
        )

    def to_datetime(self):
        return datetime(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
        )

    @classmethod
    def parse(cls, datestr):
        return Date.from_datetime(parser.parse(datestr))

    def format(self, format_str=None):
        fmt = format_str if format_str else FMT_DATETIME
        return self.strftime(fmt)

    def days_to_end_of_week(self):
        week_day = self.std_weekday()
        return 7 - week_day

    def days_from_start_of_week(self):
        return self.std_weekday() - 1

    def start_of_week(self):
        """
        This will return the preceeding Sunday at 12:00:00 am.

        """
        bow = self - timedelta(days=self.days_from_start_of_week())
        return self.from_datetime(bow).start_of_day()

    def end_of_week(self):
        """
        This will return the following Sunday at 12:00:00 am.

        """
        eow = self + timedelta(days=self.days_to_end_of_week() + 1)
        return self.from_datetime(eow).start_of_day()

    def start_of_day(self, hour=0, minute=0):
        return self.replace(hour=hour, minute=minute, second=0, microsecond=0)


class Timer:
    def __init__(self, start_time):
        if isinstance(start_time, datetime):
            self.start_time = start_time
        elif isinstance(start_time, str):
            self.start_time = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        else:
            self.start_time = start_time

    def get_lapsed_seconds(self):
        end_time = datetime.now()
        duration = end_time - self.start_time
        return duration.total_seconds()

    def get_formatted_lapsed_time(self):
        total_seconds = self.get_lapsed_seconds()
        seconds = int(total_seconds) % 60
        total_minutes = int(total_seconds) / 60
        minutes = int(total_minutes % 60)
        hours = int(int(total_minutes) / 60)
        return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)
